@@ -1,15 +1,53 @@
-from typing import Union, List
-from pathlib import Path
-import functools
 import re
 import json
+import types
+import functools
+from pathlib import Path
+from typing import Union, List
 from collections import namedtuple
 
 import pyspark.sql.functions as F
 from pyspark.sql import SparkSession, DataFrame
 from pyspark.sql.column import Column
+from pyspark.sql.types import (
+    ArrayType,
+    StringType,
+    StructType,
+    MapType,
+    BinaryType,
+    BooleanType,
+    DateType,
+    NullType,
+    NumericType,
+    TimestampType,
+    UserDefinedType,
+)
 
-from pyspark.sql.types import StructType
+
+def validate_config_path(path: Union[str, Path]):
+    """
+    Pending deprecation.
+    """
+    path = Path(path)
+
+    # validate user-supplied path
+    #   - Must end with '.config.json'
+    name_parts = path.resolve().name.split(".")
+    if name_parts[-1] != "json" or name_parts[-2] != "config":
+        raise ValueError(
+            "path to dataframe config file must end with '.config.json'\n"
+            f"path: {path.resolve()}"
+        )
+
+
+def show_by_type(sdf: DataFrame, dtype: str, *args, **kwargs):
+    """
+    .show() columns of provided dtype
+
+    Engine: Spark
+    """
+    columnList = [item[0] for item in sdf.dtypes if item[1].startswith(dtype)]
+    sdf.select(columnList).show(*args, **kwargs)
 
 
 def namedtuple_from_schema(schema: StructType, name: str):
@@ -76,9 +114,9 @@ def dict_to_map(mapping: dict) -> Column:
         if isinstance(v, (str, int, float, bool)):
             literal_list.append(F.lit(v))
         elif isinstance(v, (list, tuple, set)):
-            literal_list.append(F.array([F.lit(x) for x in v]))
+            literal_list.append(F.array(*[F.lit(x) for x in v]))
 
-    return F.create_map(literal_list)
+    return F.create_map(*literal_list)
 
 
 def map_column(col_name: str, mapping: Union[Column, dict]) -> Column:
@@ -245,6 +283,40 @@ def stage_dataframe_to_disk(
     return decorator_to_disk
 
 
+def data_template_from_schema(schema: StructType):
+    """Returns a hand-fillable template based on supplied schema."""
+    header_string = ""
+    output = "("
+    for field in schema.fields:
+        if isinstance(field.dataType, StringType):
+            output += '"String", '
+        elif isinstance(field.dataType, DateType):
+            output += 'date.fromisoformat(""), '
+        elif isinstance(field.dataType, TimestampType):
+            output += 'datetime.fromisoformat(""), '
+        elif isinstance(field.dataType, ArrayType):
+            output += '["String?"], '
+        elif isinstance(field.dataType, (MapType, StructType)):
+            output += '{"": ""}, '
+        elif isinstance(field.dataType, BinaryType):
+            output += 'b"", '
+        elif isinstance(field.dataType, BooleanType):
+            output += "True False, "
+        elif isinstance(field.dataType, NumericType):
+            output += f"{field.dataType}, "
+        elif isinstance(field.dataType, NullType):
+            output += "None, "
+        elif isinstance(field.dataType, UserDefinedType):
+            output += "UDT(), "
+        else:
+            # Something numeric
+            output += "Error, "
+        header_string += field.name + ", "
+    header_string = header_string[:-2]
+    output = output[:-2] + ")"
+    return header_string + "\n" + output
+
+
 def set_column_order(
     df: DataFrame, column_order: list, remove_unlisted: bool = False
 ) -> DataFrame:
@@ -315,27 +387,20 @@ def outer_union_corr(*dfs: DataFrame) -> DataFrame:
     distinct(). The difference between this function and unionByName() is that this
     function returns all columns from all DataFrames.
 
+    This is equivalent to chaining unionByName() with allowMissingColumns=True.
+
     Args:
         dfs: Sequence of DataFrames to union
 
     Returns:
         DataFrame
     """
-    # Identify all unique column names from all dfs
-    all_cols = set.union(*[set(i.columns) for i in dfs])
-
     for i, df in enumerate(dfs):
-        # Identify missing columns from each df
-        missing_cols = all_cols.difference(set(df.columns))
-
-        # Assign missing columns to each df with null values
-        __df = df.select("*", *[F.lit(None).alias(col) for col in missing_cols])
-
         # For first df, reassign to _df. Otherwise overwrite and union with _df
         if i == 0:
-            _df = __df
+            _df = df
         else:
-            _df = _df.unionByName(__df)
+            _df = _df.unionByName(df, allowMissingColumns=True)
 
     return _df
 
@@ -511,7 +576,9 @@ def _sql_builder(
     return sql
 
 
-def readParquetTable(self, file_path: Union[str, Path], table_name: str) -> DataFrame:
+def readParquetTable(
+    self, file_path: Union[str, Path], table_name: str = None
+) -> DataFrame:
     """
     Reads parquet files using the PySpark Table API to return a DataFrame with bucketing
     information retained. Bucketing info is only kept when writing the DataFrame using
@@ -524,6 +591,9 @@ def readParquetTable(self, file_path: Union[str, Path], table_name: str) -> Data
     Returns:
         DataFrame
     """
+    if table_name is None:
+        table_name = Path(file_path).stem
+
     if isinstance(file_path, Path):
         file_path = str(file_path)
 
@@ -566,3 +636,108 @@ def readParquetTable(self, file_path: Union[str, Path], table_name: str) -> Data
 
 DataFrame.saveParquetTable = saveParquetTable
 SparkSession.readParquetTable = readParquetTable
+
+
+def get_dataframe(obj) -> DataFrame:
+    """Attempts to return a Spark DataFrame from supplied reference."""
+    spark = SparkSession.getActiveSession()
+
+    # Cast Path to str
+    if isinstance(obj, Path):
+        obj = str(obj)
+
+    if isinstance(obj, str):
+        return spark.readParquetTable(obj)
+    if isinstance(obj, DataFrame):
+        return obj
+    if isinstance(obj, types.FunctionType):
+        return obj()
+    if isinstance(obj, types.MethodType):
+        # NOTE(Wes): Not 100% sure on this one.
+        return obj()
+    # Return whatever it is and let the end user figure it out (None, perhaps?)
+    return obj
+
+
+def melt(
+    df: DataFrame,
+    melt_cols: Union[List[str], str],
+    id_cols: Union[List[str], str, None] = None,
+    col_name: str = "col",
+    val_name: str = "val",
+) -> DataFrame:
+    """
+    Melt a DataFrame from wide to long format
+
+    Args:
+        df (DataFrame): DataFrame to melt
+        melt_cols (Union[List[str], str]): Columns to melt
+        id_cols (Union[List[str], str, None], optional): ID columns to not melt.
+            Defaults to None.
+        col_name (str, optional): Name of column containing original column names.
+            Defaults to "col".
+        val_name (str, optional): Name of column containing original values. Defaults to
+            "val".
+
+    Returns:
+        DataFrame: Melted DataFrame
+    """
+    if isinstance(id_cols, str):
+        id_cols = [id_cols]
+    if isinstance(melt_cols, str):
+        melt_cols = [melt_cols]
+
+    exploded_cols: Column = F.explode(
+        F.array(
+            *[
+                F.struct(F.lit(c).alias(col_name), F.col(c).alias(val_name))
+                for c in melt_cols
+            ]
+        )
+    ).alias("__col")
+
+    _df: DataFrame
+
+    if id_cols is None:
+        _df = df.select(exploded_cols).select(F.col("__col.*"))
+    else:
+        _df = df.select(*id_cols, exploded_cols).select(*id_cols, F.col("__col.*"))
+
+    return _df
+
+
+def merge_map_arrays(*cols: Union[str, Column]) -> Column:
+    """
+    Merge N-maps where the values are arrays by unioning them together.
+
+    Args:
+        cols (Union[str, Column]): List of columns of type Map(Any, Array) to combine
+
+    Returns:
+        Column: Merged maps of type Map(Any, Array)
+    """
+
+    if len(cols) < 2:
+        raise ValueError(
+            "Not enough values passed. merge_map_arrays must require at least 2 cols"
+        )
+
+    for i, col in enumerate(cols):
+        # Ensure columns do not contain nulls--must be an empty map
+        col = F.coalesce(col, F.create_map())
+
+        if i == 0:
+            merge_col = col
+        else:
+            # Merge maps based on key
+            # Values are arrays and will be unioned together
+            # map_zip_with defaults to null if missing key--coalesce with an empty array
+            merge_col = F.map_zip_with(
+                merge_col,
+                col,
+                lambda k, v1, v2: F.array_union(
+                    F.coalesce(v1, F.array()), F.coalesce(v2, F.array())
+                ),
+            )
+
+    return merge_col
